#! /usr/bin/env python
# coding=UTF-8

# Paquet Debian requis : python-astropy

from astropy.io import fits
from astropy.io.fits import PrimaryHDU
from astropy.io.fits import Header
from astropy import wcs

import requests # installs with : pip install requests

import sys
import time

import os
import re
#import shutil
import tempfile
import subprocess
import getpass
import argparse
import urllib

import json

import numpy
from gwyn import WCS_PV

# Tests :
# ./get_ossos.py -o "o3o29" -j "2456778..2456779"

#TMP_FITS_FILE = '/tmp/ossos-dl.fits'
NORD_BY_NB_PV = [0,1,1,1,2,2,2,3,3,3,3]
REGEXP_PINT = re.compile('^[0-9]+$')
REGEXP_PFLOAT = re.compile('^[0-9]+(\.[0-9]+)?$')
REGEXP_FLOAT = re.compile('^-?[0-9]+(\.[0-9]+)?$')
REGEXP_ALNUM = re.compile('^[0-9A-Za-z]+$')
REGEXP_OBSDATE = re.compile('^[0-9]{4} [0-9]{2} [0-9]{2}\.[0-9]+$')
REGEXP_RA = re.compile('^[0-9]{2} [0-9]{2} [0-9]{2}\.[0-9]+[ ]*$')
REGEXP_DEC = re.compile('^[+-][0-9]{2} [0-9]{2} [0-9]{2}\.[0-9]+[ ]*$')

def invertMatrix(cd):
	dc = numpy.array([[0., 0.], [0., 0.]])
	det = cd[0][0] * cd[1][1] - cd[0][1] * cd[1][0]
	dc[0][0] =  cd[1][1] / det
	dc[1][1] =  cd[0][0] / det
	dc[0][1] = -cd[0][1] / det
	dc[1][0] = -cd[1][0] / det
	return dc

def getXY(hdu, ra, dec):
	# Prepare important records in header
	hduForWCS = Header.fromkeys(['RADESYSa','CTYPE1','CTYPE2','CRVAL1', 'CRVAL2', 'CRPIX1', 'CRPIX2', 'CD1_1', 'CD1_2', 'CD2_1', 'CD2_2'])
	hduForWCS['RADESYSa'] = 'FK5'
	hduForWCS['CTYPE1'] = hdu.header['CTYPE1']
	hduForWCS['CTYPE2'] = hdu.header['CTYPE2']
	hduForWCS['CRVAL1'] = hdu.header['CRVAL1']
	hduForWCS['CRVAL2'] = hdu.header['CRVAL2']
	hduForWCS['CRPIX1'] = hdu.header['CRPIX1']
	hduForWCS['CRPIX2'] = hdu.header['CRPIX2']
	hduForWCS['CD1_1'] = hdu.header['CD1_1']
	hduForWCS['CD1_2'] = hdu.header['CD1_2']
	hduForWCS['CD2_1'] = hdu.header['CD2_1']
	hduForWCS['CD2_2'] = hdu.header['CD2_2']

	# Parse the WCS keywords
	w = wcs.WCS(hduForWCS, relax=True)
	world = numpy.array([[ra,dec]], numpy.float_)
	return w.wcs_world2pix(world, 1)


def checkValidAstrometry(a):
	#print a['objName']+'\t'+a['obs_date']

	# Check the presence of all parameters (note magnitude is optional)
	if a['objName'] == None or a['obs_date'] == None or a['ra_deg'] == None or a['ra_str'] == None or a['dec_deg'] == None or a['dec_str'] == None or a['photo_host'] == None or a['photo_id'] == None or a['photo_chip'] == None:
		print 'Required data is not available'
		return False

	# Check if it is really an OSSOS photo
	if a['photo_host'] != 'OSSOS':
		print 'Not OSSOS photo !'
		return False

	# Check input format, important for security
	if REGEXP_ALNUM.match(a['objName']) == None:
		print 'Name of object is not fully alphanumeric (or is empty)'
		return False

	if REGEXP_OBSDATE.match(a['obs_date']) == None:
		print 'Observation date is not in YYYY MM DD.ddddd format'
		return False

	if REGEXP_RA.match(a['ra_str']) == None:
		print 'RA string is not in HH MM SS.sss format'
		return False

	if REGEXP_PFLOAT.match(a['ra_deg']) == None:
		print 'RA degrees is not numeric'
		return False

	if REGEXP_DEC.match(a['dec_str']) == None:
		print 'Dec string is not in +DD MM SS.ss format'
		return False

	if REGEXP_FLOAT.match(a['dec_deg']) == None:
		print 'Dec degrees is not numeric'
		return False

	if a['obs_mag'] != None and REGEXP_FLOAT.match(a['obs_mag']) == None:
		print 'Magnitude is not numeric'
		return False

	if REGEXP_PINT.match(a['photo_id']) == None or REGEXP_PINT.match(a['photo_chip']) == None:
		print 'Photo ID or chip ID are not integer'
		return False

	return True	# All is ok !

def formatFileName(a, dst):
	dst = dst.replace('%objName', a['objName'])
	dst = dst.replace('%obs_date', a['obs_date'].strip().replace(' ','-'))
	dst = dst.replace('%ra_str', a['ra_str'].strip().replace(' ','-'))
	dst = dst.replace('%ra_deg', a['ra_deg'])
	dst = dst.replace('%dec_str', a['dec_str'].strip().replace(' ','-'))
	dst = dst.replace('%dec_deg', a['dec_deg'])
	if a['obs_mag'] != None:
		dst = dst.replace('%obs_mag', a['obs_mag'])
	else:
		dst = dst.replace('%obs_mag', '')
	dst = dst.replace('%photo_id', a['photo_id'])
	dst = dst.replace('%photo_chip', a['photo_chip'])
	return dst

def createDirsTo(dst):
	# Create directori(es)
	dst_dir = dst.rsplit('/',1)
	if len(dst_dir) == 2 and not os.path.exists(dst_dir[0]):
		print 'Create dir: '+dst_dir[0]
		os.makedirs(dst_dir[0])

def download(a, dst, cadc_user, cadc_pass, center_ra, center_dec, field):
	tmpfile = tempfile.mkstemp(prefix='tnodb_ossos_img_')[1]

	src="http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/vospace/auth/synctrans?TARGET=vos://cadc.nrc.ca~vospace/OSSOS/dbimages/"+str(a['photo_id'])+"/"+str(a['photo_id'])+"p.fits&DIRECTION=pullFromVoSpace&PROTOCOL=ivo://ivoa.net/vospace/core%23httpget&cutout=CIRCLE+ICRS+"+str(center_ra)+"+"+str(center_dec)+"+"+str(field)+"&view=cutout"
	subprocess.call(['wget', '--user='+cadc_user, '--password='+cadc_pass, '--timeout=30', '-O', tmpfile, src])

	# Selects the FITS extention where the object are
	hdulist = fits.open(tmpfile, do_not_scale_image_data=True)
	hdu = hdulist['ccd%02d' % int(a['photo_chip'])]

	# Check headers required to compute position
	lst_entetes = hdu.header.keys()
	entetes_requis = ['CRVAL1', 'CRVAL2', 'CRPIX1', 'CRPIX2', 'CD1_1', 'CD1_2', 'CD2_1', 'CD2_2']
	for entetes_test in entetes_requis:
		if not entetes_test in lst_entetes:
			print "required header "+entetes_requis+" don't exist !"
			os.unlink(tmpfile)
			return False
	
	# Get PV*_* headers if possible
	pv1 = []
	pv2 = []
	nord = 0
	low_accuracy = False
	for i in range(11):
		if ('PV1_'+str(i) in lst_entetes) and ('PV2_'+str(i) in lst_entetes):
			pv1.append(hdu.header['PV1_'+str(i)])
			pv2.append(hdu.header['PV2_'+str(i)])
			nord = NORD_BY_NB_PV[i]
		else:
			print "headers PV1_"+str(i)+" and/or PV2_"+str(i)+" don't exist : low precision"
			low_accuracy = True

	if not low_accuracy:
		# Prepare CD*_* and DC
		cd=numpy.array([
			[hdu.header['CD1_1'], hdu.header['CD1_2']],
			[hdu.header['CD2_1'], hdu.header['CD2_2']]
		])
		dc=invertMatrix(cd)

		# Compute position of the object on image using wcs
		pv=numpy.array([pv1, pv2])
		wcs_pv = WCS_PV.WCS_PV(None, hdu.header['CRVAL1'], hdu.header['CRVAL2'], hdu.header['CRPIX1'], hdu.header['CRPIX2'], cd, dc, pv, nord)
		x, y = wcs_pv.SKY2XY(float(a['ra_deg']), float(a['dec_deg']))
	else:
		# Method with lower precision if headers required for wcs in not present
		xy = getXY(hdu, float(a['ra_deg']), float(a['dec_deg']))
		x = xy[0][0]
		y = xy[0][1]
		

	# Write data on astromerty and the position of object on image to the headers
	hdu.header['TNO_OBJ']	= (a['objName'], 'TNO name')
	hdu.header['TNO_RA']	= (float(a['ra_deg']), a['ra_str'])
	hdu.header['TNO_DEC']	= (float(a['dec_deg']), a['dec_str'])
	if a['obs_mag'] != None:
		hdu.header['TNO_MAG']	= (float(a['obs_mag']), 'TNO magnitude as noted in the database')
	hdu.header['TNO_X']		= (x, 'Approximate position of the object in image')
	hdu.header['TNO_Y']		= (y, 'Approximate position of the object in image')
	hdu.header['TNO_APOS']	= (low_accuracy, 'If true, important approximation in X,Y pos.')

	# Saves the file
	outHDU = PrimaryHDU(hdu.data, hdu.header)
	outHDU.writeto(dst, clobber=True)

	os.unlink(tmpfile)
	return True


#cadc_pass = getpass.getpass(prompt='CADC password: ', stream=None)
if __name__=="__main__":
	parser = argparse.ArgumentParser(description='Download OSSOS images from CADC using the TNOdb 3 database.', add_help=False)

	serverGrp = parser.add_argument_group('Server options', 'This options concerns the way to connect to server.')
	serverGrp.add_argument('--tnodb_ws', help='URL of the REST webservice of TNOdb.', default='http://tnodb.obs-besancon.fr/ws/') #default='http://127.0.0.1/TNOdb/ws/')
	serverGrp.add_argument('--user', help='Your username for CADC login (defaults to OS\'s username : %(default)s).', default=os.getlogin())

	selGrp = parser.add_argument_group('Image selection', 'This options selects the image(s) you want. See format for selection conditions on our website for detailed format.')
	selGrp.add_argument('--all', help='Selects all objects in the database (must be set if no other criteria)', action='store_true')
	selGrp.add_argument('-o', '--obj', help='Selects object name')
	selGrp.add_argument('-j', '--jd', help='Selects date of observation in JD')
	selGrp.add_argument('--ra', help='Selects on right ascension')
	selGrp.add_argument('--dec', help='Selects on declination')
	selGrp.add_argument('--mag', help='Selects on magnitude')
	selGrp.add_argument('--obs', help='Selects on observatory code')
	selGrp.add_argument('-r', '--run', help='Selects on run')

	imgGrp = parser.add_argument_group('Image parameters', 'This options sets parameters for the images. Note the real position of the center of image also depends of borders of the image. We don\'t add pixels to force the object/required position to be centred.')
	imgGrp.add_argument('-O', '--outputfiles', help='Name of the output file, use %%item to insert a value from database, %%objName : name of object, %%obs_date for date of observation. (default : %(default)s)', default='%objName-%obs_date.fit')
	imgGrp.add_argument('--cra', help='Center image to this RA in degrees, default : to the object')
	imgGrp.add_argument('--cdec', help='Center image to this Dec in degrees, default : to the object')
	imgGrp.add_argument('-f', '--field', help='Field of view in degrees, default : %(default)s', default='0.05')

	miscGrp = parser.add_argument_group('Miscellaneous')
	miscGrp.add_argument('-h', '--help', help='Show this help message and exit.', action='store_true')
	miscGrp.add_argument('--keep', help='Don\'t download and keep files if present.', action='store_true')

	args = parser.parse_args()

	if (args.help):
		parser.print_help()
		exit(0)

	# Check some validity
	if args.cra != None and REGEXP_PFLOAT.match(args.cra) == None:
		print 'RA degrees for center of image is not a positive numeric'
		exit(4)

	if args.cdec != None and REGEXP_FLOAT.match(args.cdec) == None:
		print 'Dec degrees for center of image is not numeric'
		exit(4)

	if REGEXP_FLOAT.match(args.field) == None:
		print 'Field of view of image is not numeric'
		exit(4)

	# Prepare dict of objets
	cond = dict()
	if args.obj:
		cond['objectName'] = args.obj
	if args.jd:
		cond['odat'] = args.jd
	if args.ra:
		cond['ra'] = args.ra
	if args.dec:
		cond['dec'] = args.dec
	if args.mag:
		cond['mag'] = args.mag
	if args.obs:
		cond['obs'] = args.obs
	if args.run:
		cond['run'] = args.run

	# If no conditions --all should be set
	if len(cond) == 0 and not args.all:
		print 'No selection options.'
		print 'If you want download entire database, use \"--all\" parameter.'
		print 'You can read help with the \"--help\" parameter.'
		exit(4)
	
	# If some conditions --all should be avoided
	if len(cond) > 0 and args.all:
		print 'The \"--all\" parameter is set but some selection conditions has been set.'
		exit(4)

	# Test if database is OK
	r = requests.get(args.tnodb_ws+"/tno/availability")
	if r.status_code != 200 or r.text != '1':
		print 'TNOdb webservice is not available'
		exit(1)

	# Gets the cookie
	r = requests.get(args.tnodb_ws+"/tno/cookie")
	if r.status_code != 200:
		print 'I can\'t take a cookie from TNOdb, status_code : '+str(r.status_code)+', message : '+r.text
		exit(1)
	else:
		cookie = r.text

	# Add OSSOS to the cookie
	cadc_pass = getpass.getpass('Password for '+args.user+' at CADC: ')
	r = requests.post(args.tnodb_ws+"/tno/cookie/"+urllib.quote(cookie)+"/ossos", data={'user':args.user, 'pass':cadc_pass})
	if r.status_code != 200:
		print 'I can\'t auth for OSSOS, status_code : '+str(r.status_code)+', message : '+r.text
		exit(1)
	if r.text != '1':
		print 'I can\'t auth for OSSOS, authentication failed'
		exit(2)


	# Add condition on survey and create the JSON representation of conditions
	cond['survey'] = 'OSSOS'
	jsonCond = json.dumps(cond)
	#print jsonCond # DEBUG

	# Get astrometries
	r = requests.get(args.tnodb_ws+'/tno/data/astrometry?cond='+urllib.quote(jsonCond)+'&cookie='+urllib.quote(cookie))
	if r.status_code != 200:
		print 'I can\'t get data from TNOdb, status_code : '+str(r.status_code)+', message : '+r.text[:120]
		exit(3)
	
	data = json.loads(r.text)
	#print len(data) # DEBUG
	nbAstr = 0
	for a in data:
		nbAstr = nbAstr + 1

		# Informations about the astromerty
		print "Astrometry "+str(nbAstr)+"/"+str(len(data))
		print "obj="+a['objName']
		print "obs_date="+a['obs_date']
		print "ra_str="+a['ra_str']
		print "ra_deg="+a['ra_deg']
		print "dec_str="+a['dec_str']
		print "dec_deg="+a['dec_deg']
		print "obs_mag="+str(a['obs_mag'])
		print "photo_host="+str(a['photo_host'])
		print "id_on_host="+str(a['photo_id'])
		print "chip_id="+str(a['photo_chip'])
		print

		# Check if information about the astromerty are correct
		if not checkValidAstrometry(a):
			print "This astrometry is not well formatted, don't download !"
		else:
			# Center to object or given coordinates
			if args.cra == None:
				center_ra = a['ra_deg']
			else:
				center_ra = args.cra
			
			if args.cdec == None:
				center_dec = a['dec_deg']
			else:
				center_dec = args.cdec

			# Output file
			dst = formatFileName(a, args.outputfiles)
			createDirsTo(dst)

			# Download
			if not args.keep or not os.path.exists(dst):
				print "Download to: "+dst
				download(a, dst, args.user, cadc_pass, center_ra, center_dec, args.field)
			else:
				print "File "+dst+" exists and --keep option enabled. Don't download."
			print
	

	# Delete the cookie
	r = requests.delete(args.tnodb_ws+"/tno/cookie/"+urllib.quote(cookie))
	if r.status_code != 200:
		print 'I can\'t delete the cookie from TNOdb, status_code : '+str(r.status_code)+', message : '+r.text


	#client = SoapClient(wsdl=args.tnodb_wsdl, ns="web", trace=True)
	#client['AuthHeaderElement'] = {'username': 'mariano', 'password': 'clave'}
	#result = client.PersonaSearch(numero_documento='99999999')
